Index: constraints.md
===================================================================
--- constraints.md	(revision 196341)
+++ constraints.md	(working copy)
@@ -43,6 +43,12 @@
 (define_register_constraint "b" "ALL_REGS"
   "@internal")
 
+(define_register_constraint "u" "M16_REGS"
+  "@internal")
+
+(define_register_constraint "w" "STACK_POINTER_REG"
+  "@internal")
+
 ;; MIPS16 code always calls through a MIPS16 register; see mips_emit_call_insn
 ;; for details.
 (define_register_constraint "c" "TARGET_MIPS16 ? M16_REGS
@@ -253,6 +259,22 @@
 		 (match_test "umips_12bit_offset_address_p (op, mode)")
 		 (match_test "mips_address_insns (op, mode, false)")))
 
+(define_constraint "Zm"
+  "@internal"
+  (match_operand 0 "umips_addiusp_imm" ""))
+
+(define_constraint "Zn"
+  "@internal"
+  (match_test "umips_addiur1sp_imm_p (op)"))
+
+(define_constraint "Zo"
+  "@internal"
+  (match_operand 0 "umips_addius5_imm" ""))
+
+(define_constraint "Zp"
+  "@internal"
+  (match_test "umips_addiur2_imm_p (op)"))
+
 (define_memory_constraint "ZR"
  "@internal
   An address valid for loading/storing register exclusive"
Index: predicates.md
===================================================================
--- predicates.md	(revision 196341)
+++ predicates.md	(working copy)
@@ -122,6 +122,13 @@
 		    ? M16_REG_P (REGNO (op))
 		    : GP_REG_P (REGNO (op))")))
 
+(define_predicate "umips_addius5_imm"
+  (match_test "IN_RANGE (INTVAL (op), -8, 7)"))
+
+(define_predicate "umips_addiusp_imm"
+  (ior (match_test "IN_RANGE (INTVAL (op), 2, 257)")
+       (match_test "IN_RANGE (INTVAL (op), -258, -3)")))
+
 (define_predicate "movep_src_register"
   (and (match_code "reg")
        (ior (match_test ("IN_RANGE (REGNO (op), 2, 3)"))
@@ -131,6 +138,14 @@
   (ior (match_operand 0 "const_0_operand")
        (match_operand 0 "movep_src_register")))
 
+(define_predicate "m16_register"
+  (and (match_code "reg")
+       (match_test "M16_REG_P (REGNO (op))")))
+
+(define_predicate "sp_register"
+  (and (match_code "reg")
+       (match_test "REGNO (op) == STACK_POINTER_REGNUM")))
+
 (define_predicate "lo_operand"
   (and (match_code "reg")
        (match_test "REGNO (op) == LO_REGNUM")))
Index: mips.md
===================================================================
--- mips.md	(revision 196341)
+++ mips.md	(working copy)
@@ -412,12 +412,26 @@
 		(const_string "yes")
 		(const_string "no")))
 
+(define_attr "compression" "none,micromips,mips16"
+  (const_string "none"))
+
+(define_attr "enabled" "no,yes"
+  (if_then_else (and (eq_attr "compression" "micromips")
+		     (match_test "!TARGET_MICROMIPS"))
+	        (const_string "no")
+	        (const_string "yes")))
+  
 ;; Length of instruction in bytes.
 (define_attr "length" ""
    (cond [(and (eq_attr "extended_mips16" "yes")
 	       (match_test "TARGET_MIPS16"))
 	  (const_int 8)
 
+	  (and (eq_attr "compression" "micromips")
+	       (eq_attr "dword_mode" "no")
+	       (match_test "TARGET_MICROMIPS"))
+	  (const_int 2)
+
 	  ;; Direct microMIPS branch instructions have a range of
 	  ;; [-0x10000,0xfffe], otherwise the range is [-0x20000,0x1fffc].
 	  ;; If a branch is outside this range, we have a choice of two
@@ -1131,14 +1145,26 @@
   "")
 
 (define_insn "*add<mode>3"
-  [(set (match_operand:GPR 0 "register_operand" "=d,d")
-	(plus:GPR (match_operand:GPR 1 "register_operand" "d,d")
-		  (match_operand:GPR 2 "arith_operand" "d,Q")))]
+  [(set (match_operand:GPR 0 "register_operand" "=!u,d,!u,!w,d,d,d")
+	(plus:GPR (match_operand:GPR 1 "register_operand" "!u,d,!u,0,!w,!w,d")
+		  (match_operand:GPR 2 "arith_operand" "!u,d,Zp,Zo,Zm,Zn,Q")))]
   "!TARGET_MIPS16"
-  "@
-    <d>addu\t%0,%1,%2
-    <d>addiu\t%0,%1,%2"
+{
+  if (which_alternative == 0 
+      || which_alternative == 1)
+    return "<d>addu\t%0,%1,%2";
+  else
+    return "<d>addiu\t%0,%1,%2";
+}
   [(set_attr "alu_type" "add")
+   (set_attr_alternative "compression"
+	[(const_string "micromips")
+	 (const_string "none")
+	 (const_string "micromips")
+	 (const_string "micromips")
+	 (const_string "micromips")
+	 (const_string "micromips")
+	 (const_string "none")])
    (set_attr "mode" "<MODE>")])
 
 (define_insn "*add<mode>3_mips16"
Index: mips-protos.h
===================================================================
--- mips-protos.h	(revision 196341)
+++ mips-protos.h	(working copy)
@@ -350,6 +350,9 @@
 extern void mips_expand_vec_minmax (rtx, rtx, rtx,
 				    rtx (*) (rtx, rtx, rtx), bool);
 
+extern bool umips_match_reg_p (rtx, rtx);
+extern bool umips_addiur1sp_imm_p (rtx);
+extern bool umips_addiur2_imm_p (rtx);
 extern const char *umips_output_save_restore (bool, rtx);
 extern bool umips_save_restore_pattern_p (bool, rtx);
 extern bool umips_load_store_pair_p (bool, rtx *);
Index: mips.c
===================================================================
--- mips.c	(revision 196341)
+++ mips.c	(working copy)
@@ -622,7 +622,7 @@
   M16_REGS,	M16_REGS,	LEA_REGS,	LEA_REGS,
   LEA_REGS,	LEA_REGS,	LEA_REGS,	LEA_REGS,
   T_REG,	PIC_FN_ADDR_REG, LEA_REGS,	LEA_REGS,
-  LEA_REGS,	LEA_REGS,	LEA_REGS,	LEA_REGS,
+  LEA_REGS,	STACK_POINTER_REG, LEA_REGS,	LEA_REGS,
   FP_REGS,	FP_REGS,	FP_REGS,	FP_REGS,
   FP_REGS,	FP_REGS,	FP_REGS,	FP_REGS,
   FP_REGS,	FP_REGS,	FP_REGS,	FP_REGS,
@@ -2377,6 +2377,39 @@
   return 0;
 }
 
+/* Return true if reg1 is the same as reg0.  */
+bool
+umips_match_reg_p (rtx reg0, rtx reg1)
+{
+  if (REGNO (reg0) == REGNO (reg1))
+    return true;
+
+  return false;
+}
+
+/* Return true if imm1 is a valid microMIPS ADDIUR1SP value.  */
+bool
+umips_addiur1sp_imm_p (rtx imm1)
+{
+  return ((INTVAL (imm1) & 3) == 0 && IN_RANGE (INTVAL (imm1), 0, 252));
+}
+
+/* Return true if imm1 is a valid microMIPS ADDIUR2 value.  */
+bool umips_addiur2_imm_p (rtx imm1)
+{
+  if (INTVAL (imm1) == -1
+      || INTVAL (imm1) == 1
+      || INTVAL (imm1) == 4
+      || INTVAL (imm1) == 8 
+      || INTVAL (imm1) == 12
+      || INTVAL (imm1) == 16
+      || INTVAL (imm1) == 20
+      || INTVAL (imm1) == 24)
+    return true;
+
+  return false;
+}
+
 /* Return true if X is a legitimate address with a 12-bit offset.
    MODE is the mode of the value being accessed.  */
 
Index: mips.h
===================================================================
--- mips.h	(revision 196341)
+++ mips.h	(working copy)
@@ -1848,6 +1848,7 @@
   DSP_ACC_REGS,			/* DSP accumulator registers */
   ACC_REGS,			/* Hi/Lo and DSP accumulator registers */
   FRAME_REGS,			/* $arg and $frame */
+  STACK_POINTER_REG,		/* $sp */
   GR_AND_MD0_REGS,		/* union classes */
   GR_AND_MD1_REGS,
   GR_AND_MD_REGS,
@@ -1886,6 +1887,7 @@
   "DSP_ACC_REGS",							\
   "ACC_REGS",								\
   "FRAME_REGS",								\
+  "STACK_POINTER_REG",							\
   "GR_AND_MD0_REGS",							\
   "GR_AND_MD1_REGS",							\
   "GR_AND_MD_REGS",							\
@@ -1925,6 +1927,7 @@
   { 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x003f0000 },	/* DSP_ACC_REGS */	\
   { 0x00000000, 0x00000000, 0x00000003, 0x00000000, 0x00000000, 0x003f0000 },	/* ACC_REGS */		\
   { 0x00000000, 0x00000000, 0x00006000, 0x00000000, 0x00000000, 0x00000000 },	/* FRAME_REGS */	\
+  { 0x20000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000 },	/* STACK_POINTER_REG */	\
   { 0xffffffff, 0x00000000, 0x00000001, 0x00000000, 0x00000000, 0x00000000 },	/* GR_AND_MD0_REGS */	\
   { 0xffffffff, 0x00000000, 0x00000002, 0x00000000, 0x00000000, 0x00000000 },	/* GR_AND_MD1_REGS */	\
   { 0xffffffff, 0x00000000, 0x00000003, 0x00000000, 0x00000000, 0x00000000 },	/* GR_AND_MD_REGS */	\
